/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)event.c 1.7      04/02/21 Chris G. Davis"

/** \file src/fsm/event.c
 *  \brief FSM event code
 *  \author Chris G. Davis 
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lous/fsm.h"
#include "internal/fsm.h"

fsm_event_t* fsm_event_alloc(char* tag, char* nextstate) {
  fsm_event_t* retval = NULL;

  LDFunc();
  LDParam(tag, == NULL, "%s", NULL);
  LDParam(nextstate, == NULL, "%s", NULL);

  Lmalloc( retval, sizeof( fsm_event_t ), NULL);

  retval->params = hash_alloc(8, 0,  
			      HASH_MULTIPLEX | HASH_SPLAY | LIST_KEEPFIFO,
			      NULL, NULL, NULL );

  if( retval->params == NULL ) {
    free( retval );
    LError("Unable to alloc");
    retval = NULL;
  } else {
    retval->num_params = 0;
    Lstrdup(retval->tag, tag, NULL, 
	    hash_destroy( retval->params); free( retval ); );
    Lstrdup(retval->target_tag, nextstate, NULL, 
	    hash_destroy( retval->params); free( retval->tag); free( retval ); );

    retval->target = NULL;
  }

  return retval;
}

void fsm_event_add_param(fsm_event_t* event, char* trigger, s32_t val) {
  fsm_event_param_t* param;

  LDFunc();

  LDParam(event, == NULL, "%p");
  LDParam(trigger, == NULL, "%s");
  LDParam(val, != val, "%d");

  Lmalloc(param, sizeof( fsm_event_param_t), );

  Lstrdup(param->trigger_tag, trigger, , 
	  free( param ); );

  param->val = val;

  if( hash_insert( event->params, trigger, param ) != 0 ) {
    LError("Unable to insert param");
    free( param->trigger_tag );
    free( param );
  } else {
    event->num_params++;
  }
}

s32_t fsm_event_fixup(fsm_t* fsm, fsm_state_t* state, fsm_event_t* event ) {
  s32_t retval = 0;
  fsm_event_param_t* param;
  iterator_t* iter;

  if( event->target == NULL ) {
    if( (event->target = hash_lookup(fsm->states, event->target_tag) ) == NULL ) {
      LError("Unable to find target %s for event %s, state %s\n",
	      event->target_tag, event->tag, state->tag);
      errno = EINVAL;
      return -1;
    }
    retval = 1;
  }

  iter = iterator_alloc( event->params );
  while( (param = iterator_getnext( iter )) != NULL ) {
    if( hash_lookup(fsm->triggers, param->trigger_tag) == NULL ) {
      LError("Unable to find trigger %s in event %s, state %s\n",
	      param->trigger_tag, event->tag, state->tag);
      errno = EINVAL;
      return -1;
    }
  }
  iterator_destroy( iter );

  return retval;
}

s32_t fsm_event_matches(fsm_event_t* event, hash_t* trigger_state) {
  fsm_event_param_t* param;
  fsm_event_param_t* value;
  s32_t retval = 0;
  iterator_t* iter;

  iter = iterator_alloc( event->params );
  while( ( param = iterator_getnext( iter )) != NULL ) {
    value = hash_lookup(trigger_state,  param->trigger_tag );

    if( param->val != value->val ) {
      retval++;
    }
  }
  iterator_destroy( iter );

  return retval;
}

void fsm_event_display(fsm_event_t* event, s32_t flags) {
  fsm_event_param_t* param;
  iterator_t* iter;


  printf("  Event: %-30s Next: %-30s\n", event->tag, event->target_tag);

  iter = iterator_alloc( event->params );
  while( ( param = iterator_getnext( iter )) != NULL ) {
    printf("    %s = %d\n", param->trigger_tag, param->val);
  }

  iterator_destroy( iter );
}
