/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)fsm.c 1.10      04/02/21 Chris G. Davis"

/** \file src/fsm/fsm.c
 *  \brief FSM general code
 *  \author Chris G. Davis 
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lous/lous.h"
#include "lous/logging.h"
#include "lous/fsm.h"
#include "internal/fsm.h"

fsm_t* fsm_alloc(fsm_state_t* initial, void (*callback)(fsm_t*, char*, char*, char*)) {
  fsm_t* retval = NULL;

  LDFunc();
  LDParam(initial, == NULL, "%p", NULL);

  Lmalloc( retval, sizeof( fsm_t ), NULL);

  retval->states = hash_alloc(16, 0, 
			      HASH_MULTIPLEX | HASH_SPLAY | LIST_KEEPFIFO,
			      NULL, NULL, NULL );
  if( retval->states == NULL ) {
    LError("Unable to malloc states");
    fsm_destroy( retval );
    return NULL;
  }

  retval->triggers = hash_alloc(16, 0, 
				HASH_MULTIPLEX | HASH_SPLAY | LIST_KEEPFIFO,
				NULL, NULL, NULL );
  if( retval->triggers == NULL ) {
    LError("Unable to malloc triggers");
    fsm_destroy( retval );
    return NULL;
  }

  retval->users = hash_alloc( 16, 0,
				HASH_MULTIPLEX | HASH_SPLAY | LIST_KEEPFIFO,
				NULL, NULL, NULL );
  if( retval->users == NULL ) {
    LError("Unable to malloc users");
    fsm_destroy( retval );
    return NULL;
  }

  retval->initial = initial;
  if( hash_insert(retval->states, initial->tag, initial) != 0 ) {
    LError("Unable to add initial state");
    fsm_destroy( retval );
    return NULL;
  }
  retval->num_states = 1;
  retval->num_triggers = 0;
  retval->num_users = 0;
  retval->lock = false;
  retval->callback = callback;
  
  return retval;
}

void fsm_destroy(fsm_t* fsm) {
  LDFunc();
  LDParam(fsm, == NULL, "%p");

  if( fsm->users ) {
    hash_destroy( fsm->users );
  }

  if( fsm->triggers ) {
    hash_destroy( fsm->triggers );
  }

  if( fsm->states ) {
    hash_destroy( fsm->states );
  }

  free( fsm );
}

s32_t fsm_lock(fsm_t* fsm) {
  s32_t retval = 1;
  fsm_state_t* state;
  iterator_t* iter;

  LDFunc();
  LDParam(fsm, == NULL, "%p", -1);
  
  /*
   * basically, loop through and find all references and setup the pointers
   */
  while( retval > 0 ) {
    retval = 0;

    iter = iterator_alloc(fsm->states);
    while( ( state = iterator_getnext( iter )) != NULL ) {
      retval += fsm_state_fixup( fsm, state );
    }
    iterator_destroy( iter );

  }

  if( retval == 0 ) {
    fsm->lock = true;
  }

  return retval;
}

void fsm_run(fsm_t* fsm) {
  fsm_user_t* user;
  iterator_t* iter;

  LDFunc();
  LDParam(fsm, == NULL, "%p");

  /*
    basic process for running fsm's:

    - loop through all users, run call back on current state
    - loop through all users and update current state based on triggers
    
  */

  iter = iterator_alloc( fsm->users );
  while( ( user = iterator_getnext( iter )) != NULL ) {
    fsm_user_run(fsm, user);
  }
  iterator_destroy( iter );

}

void fsm_set_trigger(fsm_t* fsm, char* user_tag, char* tag, s32_t val) {
  fsm_user_t* user;

  LDFunc();
  LDParam(fsm, == NULL, "%p");
  LDParam(user_tag, == NULL, "%s");
  LDParam(tag, == NULL, "%s");
  LDParam(val, != val, "%d");

  user = hash_lookup( fsm->users, user_tag);

  if( user == NULL ) {
    LError("user not found %s", user_tag);
    errno = ENOENT;
  } else {
    fsm_user_set_trigger(user, tag, val);
  }
}

s32_t fsm_get_trigger(fsm_t* fsm, char* user_tag, char* tag) {
  fsm_user_t* user;

  LDFunc();
  LDParam(fsm, == NULL, "%p", -1);
  LDParam(user_tag, == NULL, "%s", -1);
  LDParam(tag, == NULL, "%s", -1);

  user = hash_lookup( fsm->users, user_tag);

  if( user == NULL ) {
    LError("User not found %s", user_tag);
    errno = ENOENT;
    return 0;
  } else {
    return fsm_user_get_trigger(user, tag);
  }
}

void fsm_inc_trigger(fsm_t* fsm, char* user_tag, char* tag, s32_t val) {
  fsm_user_t* user;

  LDFunc();
  LDParam(fsm, == NULL, "%p");
  LDParam(user_tag, == NULL, "%s");
  LDParam(tag, == NULL, "%s");
  LDParam(val, != val, "%d");

  user = hash_lookup( fsm->users, user_tag);

  if( user == NULL ) {
    LError("User not found %s", user_tag);
    errno = ENOENT;
  } else {
    fsm_user_inc_trigger(user, tag, val);
  }
}

void fsm_force_state(fsm_t* fsm, char* user_tag, char* tag) {
  fsm_user_t* user;

  LDFunc();
  LDParam(fsm, == NULL, "%p");
  LDParam(user_tag, == NULL, "%s");
  LDParam(tag, == NULL, "%s");

  user = hash_lookup( fsm->users, user_tag);

  if( user == NULL ) {
    LError("User not found %s", user_tag);
    errno = ENOENT;
  } else {
    fsm_user_force_state(user, tag);
  }
}

void fsm_display_tag(fsm_t* fsm, char* tag) {
  fsm_user_t* user;

  user = hash_lookup( fsm->users, tag);

  if( user == NULL ) {
    LError("User not found %s", tag);
    errno = ENOENT;
  } else {
    fsm_user_display(user);
  }
}

void fsm_display_fsm(fsm_t* fsm, s32_t flags) {
  fsm_state_t* state;
  iterator_t* iter;

  printf("FSM\n");
  printf("----------------------\n");
  printf("States: %d\n", fsm->num_states);
  printf("Triggers: %d\n", fsm->num_triggers);
  printf("Users: %d\n", fsm->num_users);
  printf("Locked: %s\n", (fsm->lock == true)?"Yes":"No");
  printf("\n");

  iter = iterator_alloc( fsm->states );
  while( ( state = iterator_getnext( iter )) != NULL ) {
    fsm_state_display(state, flags);
  }
  iterator_destroy( iter );

}

