/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)state.c 1.6      04/02/21 Chris G. Davis"

/** \file src/fsm/state.c
 *  \brief FSM State code
 *  \author Chris G. Davis 
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lous/lous.h"
#include "lous/logging.h"
#include "lous/fsm.h"
#include "internal/fsm.h"

fsm_state_t* fsm_state_alloc(char* tag, void (*callback)(fsm_t*, char*), 
			     s32_t flags, s32_t val) {
  fsm_state_t* retval = NULL;

  LDFunc();
  LDParam(tag, == NULL, "%s", NULL);

  Lmalloc(retval, sizeof( fsm_state_t ), NULL);

  Lstrdup(retval->tag, tag, NULL, free( retval ); );

  retval->events = hash_alloc(8, 0,
			      HASH_MULTIPLEX | HASH_SPLAY | LIST_KEEPFIFO,
			      NULL, NULL, NULL );
  
  if( retval->events == NULL ) {
    LError("Unable to malloc");
    free( retval->tag );
    free( retval );
    errno = ENOMEM;
    return NULL;
  }

  retval->num_events = 0;
  retval->callback = callback;
  retval->flags = flags;
  retval->val = val;

  return retval;
}

void fsm_state_add(fsm_t* fsm, fsm_state_t* state ) {

  LDFunc();
  LDParam(fsm, == NULL, "%p");
  LDParam(state, == NULL, "%p");

  if( fsm->lock == true ) {
    LError("fsm is locked");
    errno = EPERM;
    return;
  }

  if( hash_insert( fsm->states, state->tag, state ) != 0 ) {
    LError("Unable to insert state");
  } else {
    fsm->num_states++;
  }

}

void fsm_state_add_event(fsm_state_t* state, fsm_event_t* event) {

  LDFunc();
  LDParam(state, == NULL, "%p");
  LDParam(event, == NULL, "%p");

  if( hash_insert( state->events, event->tag, event) != 0 ) {
    LError("Unable to insert state");
  } else {
    state->num_events++;
  }

}

fsm_state_t* fsm_state_find(fsm_t* fsm, char* tag) {

  LDFunc();
  LDParam(fsm, == NULL, "%p", NULL);
  LDParam(tag, == NULL, "%s", NULL);

  return hash_lookup(fsm->states, tag);
}

s32_t fsm_state_fixup(fsm_t* fsm, fsm_state_t* state ) {
  s32_t retval = 0;
  fsm_event_t* event;
  iterator_t* iter;

  LDFunc();
  LDParam(fsm, == NULL, "%p", -1 );
  LDParam(state, == NULL, "%p", -1 );

  if( fsm->lock == true ) {
    LError("fsm is locked");
    errno = EPERM;
    return -1;
  }

  iter = iterator_alloc( state->events );
  while( ( event = iterator_getnext( iter )) != NULL ) {
    retval+= fsm_event_fixup(fsm, state, event);
  }
  iterator_destroy( iter );

  return retval;
}

void fsm_state_display(fsm_state_t* state, s32_t flags) {
  fsm_event_t* event;
  iterator_t* iter;

  printf("State: %-30s Flags: %04X ", state->tag, state->flags);
  if( state->flags & FSM_STATE_TIMEOUT ) {
    printf("Timeout: %d ", state->val);
  }
  printf("\n");

  iter = iterator_alloc( state->events );
  while( ( event = iterator_getnext( iter )) != NULL ) {
    fsm_event_display(event, flags);
  }
  iterator_destroy( iter );

  printf("\n");
}
