/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef LOUS_SRC_INCLUDE_TOKENIZER_H
#define LOUS_SRC_INCLUDE_TOKENIZER_H

#ident "@(#)tokenizer.h 1.2      04/02/21 Chris G. Davis"

/** \file src/include/tokenizer.h
 *  \brief Break a string into individual tokens
 *  \author Chris G. Davis
 *
*/

/*
 * required lous functionality
 */
#include "lous/linkedlist.h"

struct Tokenizer;
typedef struct Tokenizer tokenizer_t;

struct Tokenizer_Response;
typedef struct Tokenizer_Response tokenizer_response_t;

/** 
 * Allocate a tokenizer 
 * @return the tokenizer structure
 */
tokenizer_t* tokenizer_alloc();

/** 
 * Destroy a tokenizer
 * @param izer the previously allocated tokenizer structure to destoy
 */
void tokenizer_destroy(tokenizer_t* izer);

/** 
 * Add a new character to the list of separator characters
 * @param izer the tokenizer structure
 * @param c the charater to add
 */
void tokenizer_add_separator(tokenizer_t* izer, char c);

/** 
 * Add a new character to the list of terminator characters
 * @param izer the tokenizer structure
 * @param c the charater to add
 */
void tokenizer_add_terminator(tokenizer_t* izer, char c);

/** 
 * Add a new character to the list of ignored characters
 * @param izer the tokenizer structure
 * @param c the charater to add
 */
void tokenizer_add_ignored(tokenizer_t* izer, char c);

/** 
 * Break this string up into tokens
 * @param izer tokenizer structure 
 * @param str string to tokenize
 * @return the response, broken into tokens
 */
tokenizer_response_t* tokenizer_tokenize(tokenizer_t* izer, char* str);

/** 
 * Destroy the tokenized response
 * @param resp the tokenized response to destroy
 */
void tokenizer_destroy_response(tokenizer_response_t* resp);

/** 
 * Get the number of tokens in the response
 * @param resp the tokenized response
 * @return number of tokens
 */
s32_t tokenizer_get_numtokens(tokenizer_response_t* resp);

/** 
 * Did the string have a terminator 
 * @param resp the tokenized response
 * @return true if the string had a terminator
 */
bool_t tokenizer_get_terminator(tokenizer_response_t* resp);

/** 
 * Is the last token a partial token (no separator or terminator)
 * @param resp the tokenized response
 * @return true if the last token is a partial token
 */
bool_t tokenizer_get_ispartial(tokenizer_response_t* resp);

/** 
 * Get an individual token
 * @param resp the tokenized response
 * @param index which token to return
 * @return the individual token (word)
 */
char* tokenizer_get_token(tokenizer_response_t* resp, s32_t index);

/** 
 * Get the preceeding string
 * @param resp the tokenized response
 * @param index which preceeding string to return
 * @return the string before the token.
 */
char* tokenizer_get_preceeding(tokenizer_response_t* resp, s32_t index);

/** 
 * Get the full string
 * @param resp the tokenized response
 * @param index which string to return
 * @return the string up to and including the token
 */
char* tokenizer_get_string(tokenizer_response_t* resp, s32_t index);

/** 
 * Get the remainder string
 * @param resp the tokenized response
 * @param index which remainder string to return
 * @return the string after the token.
 */
char* tokenizer_get_remainder(tokenizer_response_t* resp, s32_t index);


#endif /* LOUS_SRC_INCLUDE_TOKENIZER_H */
