/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)netevent.c 1.13      04/02/21 Chris G. Davis"

/** \file src/networking/netevent.c
 *  \brief Net Event Code
 *  \author Chris G. Davis 
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
//#include <net/if.h>

#include "lous/hash.h"
#include "lous/netevent.h"
#include "internal/netevent.h"

netevent_t* netevent_alloc(s32_t flags, s32_t timeout, s32_t sec, s32_t usec) {
  netevent_t* retval = NULL;

  LDFunc();

  Lmalloc(retval, sizeof( netevent_t), NULL);
  
  if( ( retval->classes = 
	hash_alloc(8, 0, HASH_MULTIPLEX | HASH_SPLAY | LIST_KEEPFIFO, 
		   NULL, NULL, NULL )) == NULL ) {
    LError("unable to allocate class hash");
    free( retval);
    retval = NULL;
    errno = ENOMEM;
  } else {
    retval->sec = sec;
    retval->usec = usec;
    retval->flags = flags;
    retval->timeout = timeout;
  }


  return retval;
}

s32_t netevent_destroy(netevent_t* event) {
  s32_t retval = -1;

  LDFunc();
  LDParam(event, == NULL, "%p", -1 );

  retval = 0;

  return retval;
}

void netevent_class_remove_dead(netevent_t* event, netevent_class_t* class) {
  netevent_user_t* user;
  struct timeval current;
  iterator_t* iter;

  gettimeofday(&current, NULL);

  iter = iterator_alloc( class->users );
  while( ( user = iterator_getnext(iter)) != NULL ) {
    /* deal with this user */
    if( ( user->timeout) && (user->flags & NEVNT_TIMEOUT)) {
      if( current.tv_sec - user->active.tv_sec > user->timeout ) {
	user->flags |= NEVNT_CLOSED;
      }
    }

    if( ( class->timeout ) && ( class->flags & NEVNT_TIMEOUT )) {
      if( current.tv_sec - user->active.tv_sec > class->timeout ) {
	user->flags |= NEVNT_CLOSED;
      }
    }

    if( ( event->timeout ) && ( event->flags & NEVNT_TIMEOUT )) {
      if( current.tv_sec - user->active.tv_sec > event->timeout ) {
	user->flags |= NEVNT_CLOSED;
      }
    }
    
    if( user->flags & NEVNT_CLOSED ) {
      class->callback( user->fd, NEVNT_CLOSED, user->tag);

      hash_remove(class->users, user->tag);
    }
  }

  iterator_destroy( iter );
}

void netevent_class_handle(netevent_t* event, netevent_class_t* class,
			    fd_set* rfd, fd_set* wfd, fd_set* xfd) {
  s32_t retval;
  netevent_user_t* user;
  s32_t flags;
  iterator_t* iter;

  iter = iterator_alloc( class->users );
  while( ( user = iterator_getnext(iter)) != NULL ) {
    /* deal with this user */
    flags = 0;

    if( (user->flags & NEVNT_WATCHREAD) &&
	FD_ISSET(user->fd, rfd) ) {
      flags |= NEVNT_GOTREAD;
    }

    if( (user->flags & NEVNT_WATCHWRITE) &&
	FD_ISSET(user->fd, wfd) ) {
      flags |= NEVNT_GOTWRITE;
    }

    if( (user->flags & NEVNT_WATCHEXCEPT) &&
	FD_ISSET(user->fd, xfd) ) {
      flags |= NEVNT_GOTEXCEPT;
    }

    if( flags != 0 ) {
      if( user->flags & NEVNT_IDLETIMEOUT ||
	  class->flags & NEVNT_IDLETIMEOUT ||
	  event->flags & NEVNT_IDLETIMEOUT ) {
	gettimeofday(&user->active, NULL);
      }

      retval = class->callback( user->fd, flags, user->tag);

      /*
       * a non-zero return will close the connection
       */
      if( ( flags & NEVNT_CLOSED) || 
	  ( retval == -1 )) {
	user->flags = NEVNT_CLOSED;
      }
    }
  }

  iterator_destroy( iter );
}

void netevent_class_prepare(netevent_t* event, netevent_class_t* class, 
			    fd_set* rfd, fd_set* wfd, fd_set* xfd, s32_t *maxfd) {
  netevent_user_t* user;
  iterator_t* iter;

  iter = iterator_alloc( class->users );
  while( ( user = iterator_getnext(iter)) != NULL ) {
    /* deal with this user */

    if( (user->flags & NEVNT_NONBLOCK) || 
	(class->flags & NEVNT_NONBLOCK) ||
	(event->flags & NEVNT_NONBLOCK)) {
      fcntl( user->fd, F_SETFL, O_NONBLOCK);
    }

    if( user->flags & NEVNT_WATCHREAD ) {
      FD_SET(user->fd, rfd);
    }

    if( user->flags & NEVNT_WATCHWRITE ) {
      FD_SET(user->fd, wfd);
    }

    if( user->flags & NEVNT_WATCHEXCEPT ) {
      FD_SET(user->fd, xfd);
    }

    if( user->fd > *maxfd ) {
      *maxfd = user->fd;
    }
  }

  iterator_destroy( iter );
}

s32_t netevent_run(netevent_t* event) {
  struct timeval tv;
  fd_set rfdset;
  fd_set wfdset;
  fd_set xfdset;

  netevent_class_t* class;
  iterator_t* iter;

  s32_t retval = -1;
  s32_t maxfd = -1;

  LDFunc();
  LDParam(event, == NULL, "%p", -1 );

  /* now, first, we build up the fd_sets */
  FD_ZERO(&rfdset);
  FD_ZERO(&wfdset);
  FD_ZERO(&xfdset);

  /* loop classes */
  iter = iterator_alloc( event->classes );
  while( (class = iterator_getnext(iter)) != NULL ) {
    netevent_class_prepare(event, class, &rfdset, &wfdset, &xfdset, &maxfd);
  }

  iterator_destroy( iter );

  /* prep select */
  maxfd++;
  tv.tv_sec = event->sec;
  tv.tv_usec = event->usec;

  if( (retval = select(maxfd, &rfdset, &wfdset, &xfdset, &tv)) < 0 ) {
    LError("select return negative value");
    return retval;
  }

  /* loop classes */
  iter = iterator_alloc( event->classes );
  while( (class = iterator_getnext(iter)) != NULL ) {
    netevent_class_handle(event, class, &rfdset, &wfdset, &xfdset);
  }
  iterator_destroy( iter );

  /* go back and remove all closes users */
  iter = iterator_alloc( event->classes );
  while( (class = iterator_getnext(iter)) != NULL ) {
    netevent_class_remove_dead(event, class);
  }
  iterator_destroy( iter );

  retval = 0;

  return retval;
}

netevent_class_t* netevent_addclass(netevent_t* event, char* tag, s32_t flags,
				    s32_t timeout,
                                    s32_t (*callback)(s32_t fd, s32_t flag, char* tag)) {
  netevent_class_t* retval = NULL;

  LDFunc();
  LDParam(event, == NULL, "%p", NULL );

  Lmalloc( retval, sizeof( netevent_class_t ), NULL);
  if( ( retval->users =
	hash_alloc(8, 0, HASH_MULTIPLEX | HASH_SPLAY | LIST_KEEPFIFO, 
		   NULL, NULL, NULL)) == NULL ) {
    LError("unable to allocate users hash");
    free( retval );
    retval = NULL;
    errno = ENOMEM;
  } else {
    Lstrdup(retval->tag, tag, NULL, free( retval ); );

    retval->flags = flags;
    retval->callback = callback;
    
    retval->timeout = timeout;
    hash_insert(event->classes, tag, retval);
  }
  
  return retval;
}

s32_t netevent_remclass(netevent_t* event, netevent_class_t* class) {
  s32_t retval = -1;

  LDFunc();

  return retval;
}

s32_t netevent_remclass_name(netevent_t* event, char* tag) {
  s32_t retval = -1;

  LDFunc();
  LDParam(event, == NULL, "%p", -1 );

  retval = ( hash_remove(event->classes, tag) != NULL );

  return retval;
}

netevent_user_t* netevent_adduser(netevent_class_t* class, char* tag,
				  s32_t timeout,
				  s32_t flags, s32_t fd) {
  netevent_user_t* retval = NULL;

  LDFunc();
  LDParam(class, == NULL, "%p", NULL);

  Lmalloc( retval, sizeof( netevent_user_t ), NULL);

  Lstrdup(retval->tag,  tag, NULL, free( retval ); );

  retval->fd = fd;
  retval->flags = flags;
  retval->timeout = timeout;
  gettimeofday(&retval->active, NULL);
  
  hash_insert(class->users, tag, retval);

  return retval;
}

s32_t netevent_remuser(netevent_class_t* class, netevent_user_t* user) {
  s32_t retval = -1;

  LDFunc();

  errno = 0;

  return retval;
}

s32_t netevent_remuser_name(netevent_class_t* class, char* tag) {
  s32_t retval = -1;
  netevent_user_t* user;

  LDFunc();
  LDParam(class, == NULL, "%p", -1 );

  user = hash_remove(class->users, tag );
  if( user != NULL ) {
    if( user->tag ) {
      free( user->tag );
    }
    free( user );
    retval = 0;
  }

  return retval;
}


