/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)hash.c 1.8      04/02/16 Chris G. Davis"

/** \file src/structures/hash.c
 *  \brief Hash table code
 *  \author Chris G. Davis
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lous/hash.h"
#include "internal/hash.h"

hash_t* hash_alloc(s32_t buckets, s32_t keylen, s32_t flags,
                   s32_t (*hash_func)(void*, s32_t, s32_t),
                   comp_t (*comp_func)(void*, void*, s32_t),
                   void (*free_func)(void*)) {
  hash_t* newhash = NULL;

  LDFunc();
  LDParam(buckets, < 1, "%d", NULL);

  Lmalloc( newhash, sizeof( hash_t), NULL );

  LDebug(LD_func, LL_verbose, "new hash %p", newhash);
    
  newhash->buckets = 1;
  while (newhash->buckets < buckets ) {
    newhash->buckets *=2;
  }
  
  LDebug(LD_func, LL_verbose, "set buckets to %d", newhash->buckets);
  newhash->keylen = keylen;
  newhash->num_entries = 0;
  if( hash_func != NULL) {
    newhash->hash_func = hash_func;
  } else {
    newhash->hash_func = default_hash;
  }

  if( comp_func != NULL ) {
    newhash->comp_func = comp_func;
  } else {
    newhash->comp_func = default_comp;
  }
  
  newhash->free_func = free_func;
  
  newhash->ops = &generic_ops;
  
  if( flags & HASH_MULTIPLEX ) {
    if( flags & HASH_LINKEDLIST ) {
    } else if ( flags & HASH_SPLAY ) {
      newhash->ops = &splay_ops;
    }
  }
  
  newhash->flags = flags;
  
  if( flags & LIST_ITERATOR ) {
    iterator_list_initialize(newhash);
  }

  newhash->ops->initialize(newhash);

  return newhash;
}

s32_t hash_get_buckets(hash_t* hash) {
  s32_t retval = -1;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);

  retval = hash->buckets;

  return retval;
}

void* hash_get_bucket(hash_t* hash, s32_t bucket) {
  void* retval = NULL;

  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);
  LDParam(bucket, < 0, "%d", NULL);

  if( (bucket < 0 ) && (bucket >= hash->buckets)) {
    errno = EINVAL;
  } else {
    retval = hash->entries[bucket];
  }

  return retval;
}

s32_t hash_destroy(hash_t* hash) {
  s32_t retval = -1;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);

  if( hash->ops == NULL || hash->ops->destroy == NULL ) {
    LError("operation not implemented");
    errno = EINVAL;
    return -1;
  }

  retval = hash->ops->destroy(hash);

  return retval;
}

s32_t hash_insert(hash_t* hash, void* key, void* data) {
  s32_t retval = -1;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);
  LDParam(key, == NULL, "%p", -1);
  LDParam(data, != data, "%p", -1);

  if( hash->ops == NULL || hash->ops->insert == NULL ) {
    LError("operation not implemented");
    errno = EINVAL;
    return -1;
  }

  retval = hash->ops->insert(hash, key, data);

  return retval;
}

void* hash_remove(hash_t* hash, void* key) {
  void* retval = NULL;

  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);
  LDParam(key, == NULL, "%p", NULL);

  if( hash->ops == NULL || hash->ops->remove == NULL ) {
    LError("operation not implemented");
    errno = EINVAL;
    return NULL;
  }

  retval = hash->ops->remove(hash, key);

  return retval;
}

s32_t hash_delete(hash_t* hash, void* key) {
  s32_t retval = -1;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);
  LDParam(key, == NULL, "%p", -1);

  if( hash->ops == NULL || hash->ops->delete == NULL ) {
    LError("operation not implemented");
    errno = EINVAL;
    return -1;
  }

  retval = hash->ops->delete(hash, key);

  return retval;
}

void* hash_replace(hash_t* hash, void* key, void* data) {
  void* retval = NULL;

  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);
  LDParam(key, == NULL, "%p", NULL);
  LDParam(data, != data, "%p", NULL);

  if( hash->ops == NULL || hash->ops->replace == NULL ) {
    LError("operation not implemented");
    errno = EINVAL;
    return NULL;
  }

  retval = hash->ops->replace(hash, key, data);

  return retval;
}

void* hash_lookup(hash_t* hash, void* key) {
  void* retval = NULL;

  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);
  LDParam(key, == NULL, "%p", NULL);

  if( hash->ops == NULL || hash->ops->lookup == NULL ) {
    LError("operation not implemented");
    errno = EINVAL;
    return NULL;
  }

  retval = hash->ops->lookup(hash, key);

  return retval;
}

void** hash_getkeys(hash_t* hash) {
  void** retval = NULL;

  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);

  if( hash->ops == NULL || hash->ops->getkeys == NULL ) {
    LError("operation not implemented");
    errno = EINVAL;
    return NULL;
  }

  retval = hash->ops->getkeys(hash);

  return retval;
}

s32_t hash_numentries(hash_t* hash) {
  s32_t retval = -1;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);

  if( hash->ops == NULL || hash->ops->getkeys == NULL ) {
    LError("operation not implemented");
    errno = EINVAL;
    return -1;
  }

  retval = hash->ops->numentries(hash);

  return retval;
}

