/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)linkedlist.c 1.8      04/02/21 Chris G. Davis"

/** \file src/structures/linkedlist.c
 *  \brief Linked List structure code
 *  \author Chris G. Davis 
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lous/lous.h"
#include "lous/logging.h"
#include "lous/linkedlist.h"
#include "internal/linkedlist.h"

void ll_remove_node(linkedlist_t* list, linkedlist_node_t* node ) {
  if( node->next ) {
    node->next->prev = node->prev;
  } else {
    list->end = node->prev;
  }
  
  if( node->prev ) {
    node->prev->next = node->next;
  } else {
    list->list = node->next;
  }
  list->last = NULL;
  list->num_entries--;
}

linkedlist_t* ll_alloc(s32_t flags, void (*free_func)(void*)) {
  linkedlist_t* retval=NULL;

  LDFunc();

  errno = 0;

  Lmalloc( retval, sizeof( linkedlist_t ), NULL);

  retval->free_func = free_func;
  retval->flags = flags;
  retval->list = NULL;
  retval->num_entries = 0;

  return retval;
}


s32_t ll_destroy(linkedlist_t* list) {
  s32_t retval=-1;
  linkedlist_node_t* walker;

  LDFunc();
  LDParam(list, == NULL, "%p", -1);

  while( ( walker = list->list ) != NULL ) {
    if( list->free_func ) {
      list->free_func( walker->data );
    }
    
      ll_remove_node(list, walker);
      
      free( walker );
      retval = 0;
  }

  return retval;
}

s32_t ll_insert_begin(linkedlist_t* list, void* data) {
  s32_t retval = -1;
  linkedlist_node_t* node;

  LDFunc();
  LDParam(list, == NULL, "%p", -1);

  Lmalloc( node, sizeof( linkedlist_node_t ), -1);

  node->data = data;
  
  node->prev = NULL;
  node->next = list->list;
  list->list = node;
  if( node->next == NULL ) {
    list->end = node;
  } else {
    node->next->prev = node;
  }
  
  retval = 0;

  return retval;
}

s32_t ll_insert_end(linkedlist_t* list, void* data) {
  s32_t retval = -1;
  linkedlist_node_t* node;

  LDFunc();
  LDParam(list, == NULL, "%p", -1);

  Lmalloc(node, sizeof( linkedlist_node_t ), -1 );

  node->data = data;
  
  node->prev = list->end;
  node->next = NULL;
  list->end = node;
  if( node->prev == NULL ) {
    list->list = node;
  } else {
    node->prev->next = node;
  }
  
  retval = 0;

  return retval;
}

s32_t ll_insert_here(linkedlist_t* list, void* data) {
  s32_t retval = -1;
  linkedlist_node_t* node;

  LDFunc();
  LDParam(list, == NULL, "%p", -1);

  if( list->last == NULL ) {
    errno = EINVAL;
  } else {

    Lmalloc( node, sizeof( linkedlist_node_t ), -1 );

    node->data = data;
    
    node->prev = list->last->prev;
    node->next = list->last;
    node->next->prev = node;
    
    if( node->prev != NULL ) {
      node->prev->next = node;
    } else {
      list->list = node;
    }
    
    retval = 0;
  }

  return retval;
}

void* ll_remove_begin(linkedlist_t* list) {
  void* retval = NULL;
  linkedlist_node_t* old;

  LDFunc();
  LDParam(list, == NULL, "%p", NULL);

  if( list->list != NULL ) {
    retval = list->list->data;
    
    old = list->list;
    
    ll_remove_node(list, old);
    
    free( old );
  }

  return retval;
}

void* ll_remove_end(linkedlist_t* list) {
  void* retval = NULL;
  linkedlist_node_t* old;

  LDFunc();
  LDParam(list, == NULL, "%p", NULL);

  if( list->end != NULL ) {
    retval = list->end->data;
    
    old = list->end;
    
    ll_remove_node(list, old);
    
    free( old );
  }

  return retval;
}

void* ll_remove_here(linkedlist_t* list) {
  void* retval = NULL;
  linkedlist_node_t* old;

  LDFunc();
  LDParam(list, == NULL, "%p", NULL);

  if( list->last != NULL ) {
    retval = list->last->data;
    
    old = list->last;
    
    ll_remove_node(list, old);
    
    free( old );
    
    retval = 0;
  }

  return retval;
}

void* ll_getfirst(linkedlist_t* list) {
  void* retval = NULL;

  LDFunc();
  LDParam(list, == NULL, "%p", NULL);

  list->last = list->list;
  
  if( list->last ) {
    retval = list->last->data;
  }

  return retval;
}

void* ll_getnext(linkedlist_t* list) {
  void* retval = NULL;

  LDFunc();
  LDParam(list, == NULL, "%p", NULL);

  if( list->last == NULL ) {
    retval = ll_getfirst(list);
  } else {
    if( list->last->next != NULL ) {
      list->last = list->last->next;
      
      retval = list->last->data;
    }
  }

  return retval;
}

void ll_clearnext(linkedlist_t* list) {
  LDFunc();
  LDParam(list, == NULL, "%p");

  list->last = NULL;
}
