/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)splay.c 1.11      04/02/16 Chris G. Davis"

/** \file src/structures/splay.c
 *  \brief Splay Tree code
 *  \author Chris G. Davis
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lous/hash.h"
#include "lous/splay.h"
#include "internal/splay.h"

splay_node_t* splay_node(splay_tree_t* tree, void* key, splay_node_t** node,
		splay_node_t** parent, splay_node_t** grandparent) {
  splay_node_t* retval;
  splay_node_t** next;
  splay_node_t* n;

  comp_t compval;

  n = *node;

  if( !n )
    return *parent;

  compval = tree->comp_func(key, n->key, tree->keylen);

  if( compval == equal ) {
    next = NULL;
  } else if ( compval == lessthen ) {
    next = &n->left;
  } else {
    next = &n->right;
  }

  if( next != NULL ) {
    n = splay_node(tree, key, next, node, parent);

    if( *node != n )
      return n;
  }

  if( parent == NULL ) {
    retval = n;
  } else {

    if( grandparent == NULL ) {
      if( n == (*parent)->left ) {
	*node = n->right;
	n->right = *parent;
      } else {
	*node = n->left;
	n->left = *parent;
      }

      *parent = n;
      retval = n;
    } else {

      if( (n == (*parent)->left ) &&
	  (*parent == (*grandparent)->left )) {
	retval = *parent;
	
	(*grandparent)->left = retval->right;
	retval->right = *grandparent;
	retval->left = n->right;
	n->right = retval;
	*grandparent = n;
	retval = n;
      } else if ( (n == (*parent)->right ) &&
		  (*parent == (*grandparent)->right )) {
	retval = *parent;
	(*grandparent)->right = retval->left;
	retval->left = *grandparent;
	retval->right = n->left;
	n->left = retval;
	*grandparent = n;
	retval = n;
      } else {
	if( n == (*parent)->left ) {
	  (*parent)->left = n->right;
	  n->right = *parent;
	  (*grandparent)->right = n->left;
	  n->left = *grandparent;
	  *grandparent = n;
	  retval = n;
	} else {
	  (*parent)->right = n->left;
	  n->left = *parent;
	  (*grandparent)->left = n->right;
	  n->right = *grandparent;
	  *grandparent = n;
	  retval = n;
	}
      }
    }
  }

  return retval;
}

void splay_tree(splay_tree_t* tree, void* key) {

  LDFunc();
  LDParam(tree, == NULL, "%p");

  if( tree->nodes == NULL ) 
    return;

  splay_node(tree, key, &tree->nodes, NULL, NULL);
}

void splay_delete_node(splay_node_t* node, void (*free_func)(void*)) {

  LDFunc();
  LDParam(node, == NULL, "%p");

  if( node->key) {
    free( node->key );
  }
  
  if( node->data ) {
    if( free_func ) {
      free_func( node->data );
    }
  }
  
  free(node);
}

splay_tree_t* splay_alloc(s32_t flags, s32_t keylen, 
			  comp_t (*comp_func)(void*, void*, s32_t),
			  void (*free_func)(void*)) {
  splay_tree_t* splay = NULL;

  LDFunc();

  Lmalloc( splay, sizeof( splay_tree_t ), NULL);

  splay->keylen = keylen;

  if( comp_func == NULL ) 
    splay->comp_func = default_comp;
  else 
    splay->comp_func = comp_func;
  
  splay->free_func = free_func;
  
  splay->flags = flags;
  splay->nodes = NULL;
  splay->num_entries = 0;

  if( flags & LIST_ITERATOR ) {
    iterator_list_initialize(splay);
  }
  
  return splay;
}

s32_t splay_destroy(splay_tree_t* tree) {
  s32_t retval = -1;

  LDFunc();
  LDParam(tree, == NULL, "%p", -1);

  while( tree->nodes ) {
    splay_delete(tree, tree->nodes->key );
  }

  free( tree );

  retval = 0;

  return retval;
}

s32_t splay_insert(splay_tree_t* tree, void* key, void* data) {
  s32_t retval = -1;
  comp_t compval = equal;
  splay_node_t* newnode;

  LDFunc();
  LDParam(tree, == NULL, "%p", -1);

  splay_tree(tree, key);

  if( tree->nodes ) {
    compval = tree->comp_func(tree->nodes->key, key, tree->keylen);
  }
  
  if( tree->nodes && (compval == equal) ) {
    errno = EEXIST;
  } else {
    Lmalloc( newnode, sizeof( splay_node_t ), -1 );

    if( tree->keylen == 0 ) {
      Lstrdup( newnode->key, key, -1, free(newnode); );
    } else {
      Lmalloc( newnode->key, tree->keylen, -1, free(newnode); );
      memcpy( newnode->key, key, tree->keylen);
    }

    newnode->data = data;

    if( tree->nodes != NULL ) {
      if( compval == lessthen ) {
	newnode->left = tree->nodes;
	newnode->right = newnode->left->right;
	newnode->left->right = NULL;
      } else {
	newnode->right = tree->nodes;
	newnode->left = newnode->right->left;
	newnode->right->left = NULL;
      }
    }

    if( tree->flags & LIST_ITERATOR ) {
      iterator_insert(tree, data );
    }

    tree->nodes = newnode;

    retval = 0;
  }

  return retval;
}

void* splay_remove(splay_tree_t* tree, void* key) {
  void* retval = NULL;
  comp_t compval;
  splay_node_t* left;
  splay_node_t* right;

  LDFunc();
  LDParam(tree, == NULL, "%p", NULL);

  splay_tree( tree, key );

  if( tree->nodes ) {
    compval = tree->comp_func(tree->nodes->key, key, tree->keylen);

    if( compval == equal ) {
      right = tree->nodes->right;
      left = tree->nodes->left;
      retval = tree->nodes->data;

      if( tree->flags & LIST_ITERATOR ) {
	iterator_remove(tree, retval );
      }

      splay_delete_node( tree->nodes, NULL);

      if( left ) {
	tree->nodes = left;

	if( right ) {
	  while( left->right ) {
	    left = left->right;
	  }
	  left->right = right;
	}
      } else {
	tree->nodes = right;
      }
    }
  }
  
  return retval;
}

s32_t splay_delete(splay_tree_t* tree, void* key) {
  s32_t retval = -1;
  void* odata;

  LDFunc();
  LDParam(tree, == NULL, "%p", -1);

  odata = splay_remove(tree, key);
  if( odata != NULL ) {
    if( tree->free_func != NULL ) {
      tree->free_func( odata );
    }
    retval = 0;
  }

  return retval;
}

void* splay_replace(splay_tree_t* tree, void* key, void* data) {
  void* retval = NULL;

  LDFunc();
  LDParam(tree, == NULL, "%p", NULL);

  retval = splay_remove(tree, key);
  splay_insert(tree, key, data);

  return retval;
}

splay_node_t* splay_lookup_ent(splay_tree_t* tree, void* key) {
  splay_node_t* retval = NULL;

  comp_t compval = equal;

  LDFunc();
  LDParam(tree, == NULL, "%p", NULL);

  if( tree->nodes == NULL ) {
    retval = NULL;
  } else {
    splay_tree( tree, key );
    
    if( tree->nodes ) {
      compval = tree->comp_func(tree->nodes->key, key, tree->keylen);
      
      if( compval == equal ) {
	retval = tree->nodes;
      }
    }
  }

  return retval;
}

void* splay_lookup(splay_tree_t* tree, void* key) {
  void* retval = NULL;
  splay_node_t* ret;

  LDFunc();
  LDParam(tree, == NULL, "%p", NULL);

  ret = splay_lookup_ent(tree, key );
  if( ret ) {
    errno = 0;
    retval = ret->data;
  } 

  return retval;
}

void** splay_getkeys(splay_tree_t* tree) {
  void** retval = NULL;

  LDFunc();
  LDParam(tree, == NULL, "%p", NULL);

  return retval;
}

void* splay_getpred(splay_tree_t* tree, void* key) {
  void* retval = NULL;
  comp_t compval;
  splay_node_t* node;

  LDFunc();
  LDParam(tree, == NULL, "%p", NULL);

  if( tree->nodes == NULL ) {
    retval = NULL;
  } else {
    splay_tree( tree, key );

    compval = tree->comp_func(tree->nodes->key, key, tree->keylen);

    if( compval == lessthen ) {
      retval = tree->nodes;
    } else {
      node = tree->nodes->left;
      while( node->right ) {
	node = node->right;
      }
      retval = node;
    }
  }

  return retval;
}

void* splay_getsucc(splay_tree_t* tree, void* key) {
  void* retval = NULL;
  comp_t compval;
  splay_node_t* node;

  LDFunc();
  LDParam(tree, == NULL, "%p", NULL);

  if( tree->nodes == NULL ) {
    retval = NULL;
  } else {
    splay_tree( tree, key );

    compval = tree->comp_func(tree->nodes->key, key, tree->keylen);

    if( compval == greaterthen ) {
      retval = tree->nodes;
    } else {
      node = tree->nodes->right;
      while( node->left ) {
	node = node->left;
      }
      retval = node;
    }
  }

  return retval;
}

s32_t splay_setforwards(splay_node_t* node, splay_node_t* next) {
  s32_t retval = -1;

  LDFunc();
  LDParam(node, == NULL, "%p", -1);

  node->forwards->backwards = next;
  next->forwards = node->forwards;
  node->forwards = next;
  next->backwards = node;
    
  return retval;
}

s32_t splay_setbackwards(splay_node_t* node, splay_node_t* prev) {
  s32_t retval = -1;

  LDFunc();
  LDParam(node, == NULL, "%p", -1);

  node->backwards->forwards = prev;
  prev->backwards = node->backwards;
  node->backwards = prev;
  prev->forwards = node;
    
  return retval;
}

