/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)cli.c 1.2      04/02/22 Chris G. Davis"

/** \file src/user/cli.c
 *  \brief CLI general code
 *  \author Chris G. Davis 
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>

#include "lous/cli.h"
#include "internal/cli.h"

cli_item_t* cli_item_alloc(cli_item_t* parent,
                           char* tag,
                           void (*cli_func)(cli_response_t*, tokenizer_response_t*),
                           void (*exp_func)(cli_response_t*, char*),
                           char* help_text,
                           char* reg_string,
			   s32_t level) {
  cli_item_t* retval;

  LDFunc();
  Lmalloc(retval, sizeof( cli_item_t ), NULL);

  /* 
   * why FIFO, not sorted? 
   *
   * we assume who ever is going to call us, knows what order they want. 
   */
  if( tag != NULL ) {
    Lstrdup(retval->tag, tag, NULL, cli_item_destroy(retval); );
  }

  if( help_text != NULL ) {
    Lstrdup(retval->help_text, help_text, NULL, 
	    cli_item_destroy(retval); );
  }

  if( reg_string != NULL ) {
    Lstrdup(retval->regex_string, reg_string, NULL, 
	    cli_item_destroy(retval); );

    if( regcomp(&retval->regex, retval->regex_string, REG_ICASE | REG_NOSUB | REG_EXTENDED) < 0 ) {
      LError("regex compile failed");
      cli_item_destroy(retval);
      return NULL;
    }
  }

  retval->command_func = cli_func;
  retval->expander = exp_func;
  retval->level = level;

  if( parent != NULL ) {
    cli_item_add(parent, tag, retval, NULL);
  }

  return retval;
}

void cli_item_destroy(cli_item_t* i) {
  LDFunc();
  LDParam(i, == NULL, "%p");

  if( i->items != NULL ) {
    splay_destroy( i->items );
  }

  if( i->tag != NULL ) {
    free( i->tag );
  }

  if( i->help_text != NULL ) {
    free( i->help_text );
  }

  if( i->regex_string != NULL ) {
    free( i->regex_string );
  }

  if( i->parent != NULL ) {
    /* remove me from parent */
  }

  free( i );
}

void cli_item_add(cli_item_t* parent, char* tag, cli_item_t* item, char* help_text) {
  LDFunc();
  LDParam(parent, == NULL, "%p");
  LDParam(tag, == NULL, "%s");
  LDParam(item, == NULL, "%s" );

  if( parent->items == NULL ) {
    if( ( parent->items = splay_alloc(LIST_KEEPFIFO, 0, NULL, NULL)) == NULL ) {
      LError("unable to alloc splay");
      return;
    }
  }

  if( item->tag == NULL ) {
    Lstrdup( item->tag, tag, );
  }

  if( item->help_text == NULL ) {
    Lstrdup(item->help_text, help_text,);
  }

  item->parent = parent;
  splay_insert(parent->items, tag, item );
  
}

void cli_item_display(cli_item_t* item, s32_t indent) {
  s32_t count;
  cli_item_t* sib;
  iterator_t* iter;

  for(count=0; count< indent; count++) {
    printf(" ");
  }

  if( item->tag != NULL ) {
    
    printf("%-12s [%s] %d", item->tag, item->help_text, item->level);
  } else {
    printf("root");
  }

  if( item->regex_string != NULL ) {
    printf(" '%s'", item->regex_string);
  }
  printf("\n");

  iter = iterator_alloc( item->items );
  while( ( sib = iterator_getnext( iter )) != NULL ) {
    cli_item_display(sib, indent + 2 );
  }
  iterator_destroy( iter );

}
